"""Likelihood testing scenario from https://arxiv.org/pdf/2112.03570.pdf."""

# pylint: disable = too-many-branches

from __future__ import annotations

import argparse
import importlib
import json
import logging
import os
import uuid
from collections.abc import Iterable
from datetime import datetime

import numpy as np
import sklearn
from scipy.stats import norm
from sklearn.datasets import load_breast_cancer
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from aisdc import metrics
from aisdc.attacks import report
from aisdc.attacks.attack import Attack
from aisdc.attacks.attack_report_formatter import GenerateJSONModule
from aisdc.attacks.target import Target

logging.basicConfig(level=logging.INFO)

N_SHADOW_MODELS = 100  # Number of shadow models that should be trained
EPS = 1e-16  # Used to avoid numerical issues in logit function
P_THRESH = 0.05  # default significance threshold


class DummyClassifier:
    """A Dummy Classifier to allow this code to work with get_metrics."""

    def predict(self, X_test):
        """Return an array of 1/0 depending on value in second column."""
        return 1 * (X_test[:, 1] > 0.5)

    def predict_proba(self, X_test):
        """Simply return the X_test."""
        return X_test


def _logit(p: float) -> float:
    """Return standard logit.

    Parameters
    ----------
    p : float
        value to evaluate logit at

    Returns
    -------
    li : float
        logit(p)

    Notes
    -----
    If p is close to 0 or 1, evaluating the log will result in numerical
    instabilities.  This code thresholds p at EPS and 1 - EPS where EPS
    defaults at 1e-16.
    """
    if p > 1 - EPS:  # pylint:disable=consider-using-min-builtin
        p = 1 - EPS
    p = max(p, EPS)
    return np.log(p / (1 - p))


class LIRAAttack(Attack):
    """The main LiRA Attack class."""

    # pylint: disable=too-many-instance-attributes

    def __init__(  # pylint: disable = too-many-arguments
        self,
        n_shadow_models: int = 100,
        p_thresh: float = 0.05,
        output_dir: str = "outputs_lira",
        report_name: str = "report_lira",
        training_data_filename: str = None,
        test_data_filename: str = None,
        training_preds_filename: str = None,
        test_preds_filename: str = None,
        target_model: list = None,
        target_model_hyp: dict = None,
        attack_config_json_file_name: str = None,
        n_shadow_rows_confidences_min: int = 10,
        shadow_models_fail_fast: bool = False,
        target_path: str = None,
    ) -> None:
        """Construct an object to execute a LiRA attack.

        Parameters
        ----------
        n_shadow_models : int
            number of shadow models to be trained
        p_thresh : float
            threshold to determine significance of things. For instance auc_p_value and pdif_vals
        output_dir : str
            name of the directory where outputs are stored
        report_name : str
            name of the pdf and json output reports
        training_data_filename : str
            name of the data file for the training data (in-sample)
        test_data_filename : str
            name of the file for the test data (out-of-sample)
        training_preds_filename : str
            name of the file to keep predictions of the training data (in-sample)
        test_preds_filename : str
            name of the file to keep predictions of the test data (out-of-sample)
        target_model : list
            name of the module (i.e. classification module name such as 'sklearn.ensemble') and
            attack model name (i.e. classification model name such as 'RandomForestClassifier')
        target_model_hyp : dict
            dictionary of hyper parameters for the target_model
            such as min_sample_split, min_samples_leaf etc
        attack_config_json_file_name : str
            name of the configuration file to load parameters
        n_shadow_rows_confidences_min : int
            number of minimum number of confidences calculated for
            each row in test data (out-of-sample)
        shadow_models_fail_fast : bool
            If true it stops repetitions earlier based on the given minimum
            number of confidences for each row in the test data
        target_path : str
            path to the saved trained target model and target data
        """
        super().__init__()
        self.n_shadow_models = n_shadow_models
        self.p_thresh = p_thresh
        self.output_dir = output_dir
        self.report_name = report_name
        self.training_data_filename = training_data_filename
        self.test_data_filename = test_data_filename
        self.training_preds_filename = training_preds_filename
        self.test_preds_filename = test_preds_filename
        self.target_model = target_model
        self.target_model_hyp = target_model_hyp
        self.attack_config_json_file_name = attack_config_json_file_name
        self.n_shadow_rows_confidences_min = n_shadow_rows_confidences_min
        self.shadow_models_fail_fast = shadow_models_fail_fast
        self.target_path = target_path
        if self.attack_config_json_file_name is not None:
            self._update_params_from_config_file()
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
        self.attack_metrics = None
        self.attack_failfast_shadow_models_trained = None
        self.metadata = None

    def __str__(self):
        """Return the name of the attack."""
        return "LiRA Attack"

    def attack(self, target: Target) -> None:
        """Run a LiRA attack from a Target object and a target model.

        Needs to have X_train, X_test, y_train and y_test set.

        Parameters
        ----------
        target : attacks.target.Target
            target as an instance of the Target class.
        """
        shadow_clf = sklearn.base.clone(target.model)

        target = self._check_and_update_dataset(target)

        self.run_scenario_from_preds(
            shadow_clf,
            target.X_train,
            target.y_train,
            target.model.predict_proba(target.X_train),
            target.X_test,
            target.y_test,
            target.model.predict_proba(target.X_test),
        )

    def _check_and_update_dataset(self, target: Target) -> Target:
        """Check that it is safe to use class variables to index prediction arrays.

        This has two steps:
        1. Replacing the values in y_train with their position in
        target.model.classes (will normally result in no change)
        2. Removing from the test set any rows corresponding to classes that
        are not in the training set.
        """
        logger = logging.getLogger("_check_and_update_dataset")
        y_train_new = []
        classes = list(target.model.classes_)
        for y in target.y_train:
            y_train_new.append(classes.index(y))

        target.y_train = np.array(y_train_new, int)

        logger.info(
            "new ytrain has values and counts: %s",
            f"{np.unique(target.y_train,return_counts=True)}",
        )
        ok_pos = []
        y_test_new = []
        for i, y in enumerate(target.y_test):
            if y in classes:
                ok_pos.append(i)
                y_test_new.append(classes.index(y))

        if len(y_test_new) != len(target.X_test):
            target.X_test = target.X_test[ok_pos, :]
        target.y_test = np.array(y_test_new, int)
        logger.info(
            "new ytest has values and counts: %s",
            f"{np.unique(target.y_test,return_counts=True)}",
        )

        return target

    def run_scenario_from_preds(  # pylint: disable = too-many-statements, too-many-arguments, too-many-locals
        self,
        shadow_clf: sklearn.base.BaseEstimator,
        X_target_train: Iterable[float],
        y_target_train: Iterable[float],
        target_train_preds: Iterable[float],
        X_shadow_train: Iterable[float],
        y_shadow_train: Iterable[float],
        shadow_train_preds: Iterable[float],
    ) -> tuple[np.ndarray, np.ndarray, sklearn.base.BaseEstimator]:
        """Run the likelihood test, using the "offline" version.

        See p.6 (top of second column) for details.

        Parameters
        ----------
        shadow_clf : sklearn.Model
            An sklearn classifier that will be trained to form the shadow model.
            All hyper-parameters should have been set.
        X_target_train : np.ndarray
            Data that was used to train the target model
        y_target_train : np.ndarray
            Labels that were used to train the target model
        target_train_preds : np.ndarray
            Array of predictions produced by the target model on the training data
        X_shadow_train : np.ndarray
            Data that will be used to train the shadow models
        y_shadow_train : np.ndarray
            Labels that will be used to train the shadow model
        shadow_train_preds : np.ndarray
            Array of predictions produced by the target model on the shadow data

        Returns
        -------
        mia_scores : np.ndarray
            Attack probabilities of belonging to the training set or not
        mia_labels : np.ndarray
            True labels of belonging to the training set or not
        mia_cls : DummyClassifier
            A DummyClassifier that directly returns the scores for compatibility with code
            in metrics.py

        Examples
        --------
        >>> X, y = load_breast_cancer(return_X_y=True, as_frame=False)
        >>> X_train, X_test, y_train, y_test = train_test_split(
        >>>   X, y, test_size=0.5, stratify=y
        >>> )
        >>> rf = RandomForestClassifier(min_samples_leaf=1, min_samples_split=2)
        >>> rf.fit(X_train, y_train)
        >>> mia_test_probs, mia_test_labels, mia_clf = likelihood_scenario(
        >>>     RandomForestClassifier(min_samples_leaf=1, min_samples_split=2, max_depth=10),
        >>>     X_train,
        >>>     y_train,
        >>>     rf.predict_proba(X_train),
        >>>     X_test,
        >>>     y_test,
        >>>     rf.predict_proba(X_test),
        >>>     n_shadow_models=100
        >>> )
        """
        logger = logging.getLogger("lr-scenario")

        n_train_rows, _ = X_target_train.shape
        n_shadow_rows, _ = X_shadow_train.shape
        indices = np.arange(0, n_train_rows + n_shadow_rows, 1)

        # Combine taregt and shadow train, from which to sample datasets
        combined_x_train = np.vstack((X_target_train, X_shadow_train))
        combined_y_train = np.hstack((y_target_train, y_shadow_train))

        train_row_to_confidence = {i: [] for i in range(n_train_rows)}
        shadow_row_to_confidence = {i: [] for i in range(n_shadow_rows)}

        # Train N_SHADOW_MODELS shadow models
        logger.info("Training shadow models")
        for model_idx in range(self.n_shadow_models):
            if model_idx % 10 == 0:
                logger.info("Trained %d models", model_idx)
            # Pick the indices to use for training this one
            np.random.seed(model_idx)  # Reproducibility
            these_idx = np.random.choice(indices, n_train_rows, replace=False)
            temp_x_train = combined_x_train[these_idx, :]
            temp_y_train = combined_y_train[these_idx]

            # Fit the shadow model
            shadow_clf.set_params(random_state=model_idx)
            shadow_clf.fit(temp_x_train, temp_y_train)

            # map a class to a column
            class_map = {c: i for i, c in enumerate(shadow_clf.classes_)}

            # Get the predicted probabilities on the training data
            confidences = shadow_clf.predict_proba(X_target_train)

            these_idx = set(these_idx)
            for i in range(n_train_rows):
                if i not in these_idx:
                    # If i was _not_ used for training, incorporate the logit of its confidence of
                    # being correct - TODO: should we just be taking max??
                    cl_pos = class_map.get(y_target_train[i], -1)
                    # Occasionally, the random data split will result in classes being
                    # absent from the training set. In these cases cl_pos will be -1 and
                    # we include logit(0) instead of discarding (also for the shadow data below)
                    if cl_pos >= 0:
                        train_row_to_confidence[i].append(
                            _logit(confidences[i, cl_pos])
                        )
                    else:  # pragma: no cover
                        # catch-all
                        train_row_to_confidence[i].append(_logit(0))
            # Same process for shadow data
            shadow_confidences = shadow_clf.predict_proba(X_shadow_train)
            for i in range(n_shadow_rows):
                if i + n_train_rows not in these_idx:
                    cl_pos = class_map.get(y_shadow_train[i], -1)
                    if cl_pos >= 0:
                        shadow_row_to_confidence[i].append(
                            _logit(shadow_confidences[i, cl_pos])
                        )
                    else:  # pragma: no cover
                        # catch-all
                        shadow_row_to_confidence[i].append(_logit(0))

            # Compute number of confidences for each row
            lengths_shadow_row_to_confidence = {
                key: len(value) for key, value in shadow_row_to_confidence.items()
            }
            n_shadow_confidences = self.n_shadow_rows_confidences_min
            # Stop training of shadow models when shadow_model_fail_fast is True
            # and a minimum number of confidences specified by parameter
            # (n_shadow_rows_confidences_min) are computed for each row
            if (
                not any(
                    value < n_shadow_confidences
                    for value in lengths_shadow_row_to_confidence.values()
                )
                and self.shadow_models_fail_fast
            ):
                break
        self.attack_failfast_shadow_models_trained = model_idx + 1
        # Do the test described in the paper in each case
        mia_scores = []
        mia_labels = []
        logger.info("Computing scores for train rows")
        for i in range(n_train_rows):
            true_score = _logit(target_train_preds[i, y_target_train[i]])
            null_scores = np.array(train_row_to_confidence[i])
            mean_null = 0.0
            var_null = 0.0
            if not np.isnan(null_scores).all():
                mean_null = np.nanmean(null_scores)  # null_scores.mean()
                var_null = np.nanvar(null_scores)  #
            var_null = max(var_null, EPS)  # var can be zero in some cases
            prob = norm.cdf(true_score, loc=mean_null, scale=np.sqrt(var_null))
            mia_scores.append([1 - prob, prob])
            mia_labels.append(1)

        logger.info("Computing scores for shadow rows")
        for i in range(n_shadow_rows):
            true_score = _logit(shadow_train_preds[i, y_shadow_train[i]])
            null_scores = np.array(shadow_row_to_confidence[i])
            mean_null = null_scores.mean()
            var_null = max(null_scores.var(), EPS)  # var can be zeros in some cases
            prob = norm.cdf(true_score, loc=mean_null, scale=np.sqrt(var_null))
            mia_scores.append([1 - prob, prob])
            mia_labels.append(0)

        mia_clf = DummyClassifier()
        logger.info("Finished scenario")

        mia_scores = np.array(mia_scores)
        mia_labels = np.array(mia_labels)
        y_pred_proba = mia_clf.predict_proba(mia_scores)
        self.attack_metrics = [metrics.get_metrics(y_pred_proba, mia_labels)]

    def example(self) -> None:
        """Run an example attack using data from sklearn.

        Generates example data, trains a classifier and tuns the attack
        """
        X, y = load_breast_cancer(return_X_y=True, as_frame=False)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.5, stratify=y
        )
        rf = RandomForestClassifier(min_samples_leaf=1, min_samples_split=2)
        rf.fit(X_train, y_train)
        self.run_scenario_from_preds(
            sklearn.base.clone(rf),
            X_train,
            y_train,
            rf.predict_proba(X_train),
            X_test,
            y_test,
            rf.predict_proba(X_test),
        )

    def _construct_metadata(self) -> None:
        """Construct the metadata object."""
        self.metadata = {}
        self.metadata["experiment_details"] = {}
        self.metadata["experiment_details"] = self.get_params()

        self.metadata["global_metrics"] = {}

        pdif = np.exp(-self.attack_metrics[0]["PDIF01"])

        self.metadata["global_metrics"]["PDIF_sig"] = (
            f"Significant at p={self.p_thresh}"
            if pdif <= self.p_thresh
            else f"Not significant at p={self.p_thresh}"
        )

        auc_p, auc_std = metrics.auc_p_val(
            self.attack_metrics[0]["AUC"],
            self.attack_metrics[0]["n_pos_test_examples"],
            self.attack_metrics[0]["n_neg_test_examples"],
        )
        self.metadata["global_metrics"]["AUC_sig"] = (
            f"Significant at p={self.p_thresh}"
            if auc_p <= self.p_thresh
            else f"Not significant at p={self.p_thresh}"
        )
        self.metadata["global_metrics"]["null_auc_3sd_range"] = (
            f"{0.5 - 3 * auc_std} -> {0.5 + 3 * auc_std}"
        )

        self.metadata["attack"] = str(self)

    def make_report(self) -> dict:
        """Create the report.

        Creates the output report. If self.args.report_name is not None, it
        will also save the information in json and pdf formats.

        Returns
        -------
        output : Dict
            Dictionary containing all attack output
        """
        logger = logging.getLogger("reporting")
        report_dest = os.path.join(self.output_dir, self.report_name)
        logger.info(
            "Starting reports, pdf report name = %s, json report name = %s",
            report_dest + ".pdf",
            report_dest + ".json",
        )
        output = {}
        output["log_id"] = str(uuid.uuid4())
        output["log_time"] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        self._construct_metadata()
        output["metadata"] = self.metadata
        output["attack_experiment_logger"] = self._get_attack_metrics_instances()

        json_attack_formatter = GenerateJSONModule(report_dest + ".json")
        json_report = report.create_json_report(output)
        json_attack_formatter.add_attack_output(json_report, "LikelihoodAttack")

        pdf_report = report.create_lr_report(output)
        report.add_output_to_pdf(report_dest, pdf_report, "LikelihoodAttack")
        logger.info(
            "Wrote pdf report to %s and json report to %s",
            report_dest + ".pdf",
            report_dest + ".json",
        )

        return output

    def _get_attack_metrics_instances(self) -> dict:
        """Construct the metadata object after attacks."""
        attack_metrics_experiment = {}
        attack_metrics_instances = {}

        for rep, _ in enumerate(self.attack_metrics):
            self.attack_metrics[rep]["n_shadow_models_trained"] = (
                self.attack_failfast_shadow_models_trained
            )
            attack_metrics_instances["instance_" + str(rep)] = self.attack_metrics[rep]

        attack_metrics_experiment["attack_instance_logger"] = attack_metrics_instances
        return attack_metrics_experiment

    def setup_example_data(self) -> None:
        """Create example data and save (including config).

        Intended to allow users to see how they would need to setup their own
        data. Generates train and test data .csv files, train and test
        predictions .csv files and a config.json file that can be used to run
        the attack from the command line.
        """
        X, y = load_breast_cancer(return_X_y=True)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.5, stratify=y
        )
        rf = RandomForestClassifier(min_samples_split=2, min_samples_leaf=1)
        rf.fit(X_train, y_train)
        train_data = np.hstack((X_train, y_train[:, None]))
        np.savetxt("train_data.csv", train_data, delimiter=",")

        test_data = np.hstack((X_test, y_test[:, None]))
        np.savetxt("test_data.csv", test_data, delimiter=",")

        train_preds = rf.predict_proba(X_train)
        test_preds = rf.predict_proba(X_test)
        np.savetxt("train_preds.csv", train_preds, delimiter=",")
        np.savetxt("test_preds.csv", test_preds, delimiter=",")

        config = {
            "training_data_filename": "train_data.csv",
            "test_data_filename": "test_data.csv",
            "training_preds_filename": "train_preds.csv",
            "test_preds_filename": "test_preds.csv",
            "target_model": ["sklearn.ensemble", "RandomForestClassifier"],
            "target_model_hyp": {"min_samples_split": 2, "min_samples_leaf": 1},
        }

        with open("config.json", "w", encoding="utf-8") as f:
            f.write(json.dumps(config))

    def attack_from_config(self) -> None:  # pylint: disable = too-many-locals
        """Run an attack based on the args parsed from the command line."""
        logger = logging.getLogger("run-attack")
        logger.info("Loading training data csv from %s", self.training_data_filename)
        training_data = np.loadtxt(self.training_data_filename, delimiter=",")
        X_train = training_data[:, :-1]
        y_train = training_data[:, -1].flatten().astype(int)
        logger.info("Loaded %d rows", len(X_train))

        logger.info("Loading test data csv from %s", self.test_data_filename)
        test_data = np.loadtxt(self.test_data_filename, delimiter=",")
        X_test = test_data[:, :-1]
        y_test = test_data[:, -1].flatten().astype(int)
        logger.info("Loaded %d rows", len(X_test))

        logger.info("Loading train predictions form %s", self.training_preds_filename)
        train_preds = np.loadtxt(self.training_preds_filename, delimiter=",")
        assert len(train_preds) == len(X_train)

        logger.info("Loading test predictions form %s", self.test_preds_filename)
        test_preds = np.loadtxt(self.test_preds_filename, delimiter=",")
        assert len(test_preds) == len(X_test)
        if self.target_model is None:
            raise ValueError("Target model cannot be None")
        if self.target_model_hyp is None:
            raise ValueError("Target model hyperparameters cannot be None")
        clf_module_name, clf_class_name = self.target_model
        module = importlib.import_module(clf_module_name)
        clf_class = getattr(module, clf_class_name)
        clf_params = self.target_model_hyp
        clf = clf_class(**clf_params)
        logger.info("Created model: %s", str(clf))
        self.run_scenario_from_preds(
            clf, X_train, y_train, train_preds, X_test, y_test, test_preds
        )
        logger.info("Computing metrics")


# Methods invoked by command line script
def _setup_example_data(args):
    """Call the methods to setup some example data."""
    attack_obj = LIRAAttack(
        n_shadow_models=args.n_shadow_models,
        n_shadow_rows_confidences_min=args.n_shadow_rows_confidences_min,
        output_dir=args.output_dir,
        report_name=args.report_name,
        p_thresh=args.p_thresh,
        shadow_models_fail_fast=args.shadow_models_fail_fast,
    )
    attack_obj.setup_example_data()


def _example(args):
    """Call the methods to run an example."""
    attack_obj = LIRAAttack(
        n_shadow_models=args.n_shadow_models,
        n_shadow_rows_confidences_min=args.n_shadow_rows_confidences_min,
        output_dir=args.output_dir,
        report_name=args.report_name,
        p_thresh=args.p_thresh,
        shadow_models_fail_fast=args.shadow_models_fail_fast,
    )
    attack_obj.example()
    attack_obj.make_report()


def _run_attack(args):
    """Run a command line attack based on saved files described in .json file."""
    attack_obj = LIRAAttack(
        n_shadow_models=args.n_shadow_models,
        n_shadow_rows_confidences_min=args.n_shadow_rows_confidences_min,
        p_thresh=args.p_thresh,
        output_dir=args.output_dir,
        report_name=args.report_name,
        shadow_models_fail_fast=args.shadow_models_fail_fast,
        attack_config_json_file_name=args.attack_config_json_file_name,
    )
    attack_obj.attack_from_config()
    attack_obj.make_report()


def _run_attack_from_configfile(args):
    """Run a command line attack based on saved files described in .json file."""
    attack_obj = LIRAAttack(
        attack_config_json_file_name=args.attack_config_json_file_name,
        target_path=str(args.target_path),
    )
    print(args.attack_config_json_file_name)
    target = Target()
    target.load(attack_obj.target_path)
    attack_obj.attack(target)
    attack_obj.make_report()


def main():
    """Parse args and invoke relevant code."""
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "-s",
        "--n-shadow-models",
        type=int,
        required=False,
        default=N_SHADOW_MODELS,
        action="store",
        dest="n_shadow_models",
        help=("The number of shadow models to train (default = %(default)d)"),
    )

    parser.add_argument(
        "--n-shadow-rows-confidences-min",
        type=int,
        action="store",
        dest="n_shadow_rows_confidences_min",
        default=10,
        required=False,
        help=(
            """Number of confidences against rows in shadow data from the shadow models
            and works when --shadow-models-fail-fast = True. Default = %(default)d"""
        ),
    )

    parser.add_argument(
        "--output-dir",
        type=str,
        action="store",
        dest="output_dir",
        default="output_lira",
        required=False,
        help=("Directory name where output files are stored. Default = %(default)s."),
    )

    parser.add_argument(
        "--report-name",
        type=str,
        action="store",
        dest="report_name",
        default="report_lira",
        required=False,
        help=(
            """Filename for the pdf and json output reports. Default = %(default)s.
            Code will append .pdf and .json"""
        ),
    )

    parser.add_argument(
        "-p",
        "--p-thresh",
        type=float,
        action="store",
        dest="p_thresh",
        required=False,
        default=P_THRESH,
        help=("Significance threshold for p-value comparisons. Default = %(default)f"),
    )

    parser.add_argument(
        "--shadow-models-fail-fast",
        action="store_true",
        required=False,
        dest="shadow_models_fail_fast",
        help=(
            """To stop training shadow models early based on minimum number of
            confidences across all rows (--n-shadow-rows-confidences-min)
            in the shadow data. Default = %(default)s"""
        ),
    )

    subparsers = parser.add_subparsers()
    example_parser = subparsers.add_parser("run-example", parents=[parser])
    example_parser.set_defaults(func=_example)

    attack_parser = subparsers.add_parser("run-attack", parents=[parser])
    attack_parser.add_argument(
        "-j",
        "--attack-config-json-file-name",
        action="store",
        required=True,
        dest="attack_config_json_file_name",
        type=str,
        help=(
            "Name of the .json file containing details for the run. Default = %(default)s"
        ),
    )
    attack_parser.set_defaults(func=_run_attack)

    attack_parser_config = subparsers.add_parser("run-attack-from-configfile")
    attack_parser_config.add_argument(
        "-j",
        "--attack-config-json-file-name",
        action="store",
        required=True,
        dest="attack_config_json_file_name",
        type=str,
        default="config_lira_cmd.json",
        help=(
            "Name of the .json file containing details for the run. Default = %(default)s"
        ),
    )

    attack_parser_config.add_argument(
        "-t",
        "--attack-target-folder-path",
        action="store",
        required=True,
        dest="target_path",
        type=str,
        default="lira_target",
        help=(
            """Name of the target directory to load the trained target model and the target data.
            Default = %(default)s"""
        ),
    )

    attack_parser_config.set_defaults(func=_run_attack_from_configfile)

    example_data_parser = subparsers.add_parser("setup-example-data")
    example_data_parser.set_defaults(func=_setup_example_data)

    args = parser.parse_args()
    try:
        args.func(args)
    except AttributeError as e:  # pragma:no cover
        print(e)
        print("Invalid command. Try --help to get more details")


if __name__ == "__main__":  # pragma:no cover
    main()



"""Structural attacks.

Runs a number of 'static' structural attacks based on:
(i) the target model's properties;
(ii) the TREs risk appetite as applied to tables and standard regressions.
"""

from __future__ import annotations

import argparse
import logging
import os
import uuid
from datetime import datetime

import numpy as np
from acro import ACRO

# tree-based model types currently supported
from sklearn.base import BaseEstimator
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.tree import DecisionTreeClassifier
from xgboost.sklearn import XGBClassifier

from aisdc.attacks import report
from aisdc.attacks.attack import Attack
from aisdc.attacks.attack_report_formatter import GenerateJSONModule
from aisdc.attacks.target import Target

logging.basicConfig(level=logging.INFO)


def get_unnecessary_risk(model: BaseEstimator) -> bool:
    """Check whether model hyperparameters are in the top 20% most risky.

    This check is designed to assess whether a model is likely to be
    **unnecessarily** risky, i.e., whether it is highly likely that a different
    combination of hyper-parameters would have led to model with similar or
    better accuracy on the task but with lower membership inference risk.

    The rules applied from an experimental study using a grid search in which:
    - max_features was one-hot encoded from the set [None, log2, sqrt]
    - splitter was encoded using 0=best, 1=random

    The target models created were then subject to membership inference attacks
    (MIA) and the hyper-param combinations rank-ordered according to MIA AUC.
    Then a decision tree trained to recognise whether hyper-params combintions
    were in the 20% most risky.  The rules below were extracted from that tree
    for the 'least risky' nodes.

    Notes
    -----
    Returns True if high risk, otherwise False.
    """
    if not isinstance(
        model, (DecisionTreeClassifier, RandomForestClassifier, XGBClassifier)
    ):
        return 0  # no experimental evidence to support rejection

    unnecessary_risk = 0
    max_depth = float(model.max_depth) if model.max_depth else 500

    # pylint:disable=chained-comparison,too-many-boolean-expressions
    if isinstance(model, DecisionTreeClassifier):
        max_features = model.max_features
        min_samples_leaf = model.min_samples_leaf
        min_samples_split = model.min_samples_split
        splitter = model.splitter
        if (
            (max_depth > 7.5 and min_samples_leaf <= 7.5 and min_samples_split <= 15)
            or (
                splitter == "best"
                and max_depth > 7.5
                and min_samples_leaf <= 7.5
                and min_samples_split > 15
            )
            or (
                splitter == "best"
                and max_depth > 7.5
                and 7.5 < min_samples_leaf <= 15
                and max_features is None
            )
            or (
                splitter == "best"
                and 3.5 < max_depth <= 7.5
                and max_features is None
                and min_samples_leaf <= 7.5
            )
            or (
                splitter == "random"
                and max_depth > 7.5
                and min_samples_leaf <= 7.5
                and max_features is None
            )
        ):
            unnecessary_risk = 1
    elif isinstance(model, RandomForestClassifier):
        n_estimators = model.n_estimators
        max_features = model.max_features
        min_samples_leaf = model.min_samples_leaf
        min_samples_split = model.min_samples_split
        if (
            (max_depth > 3.5 and n_estimators > 35 and max_features is not None)
            or (
                max_depth > 3.5
                and n_estimators > 35
                and min_samples_split <= 15
                and max_features is None
                and model.bootstrap
            )
            or (
                max_depth > 7.5
                and 15 < n_estimators <= 35
                and min_samples_leaf <= 15
                and not model.bootstrap
            )
        ):
            unnecessary_risk = 1

    elif isinstance(model, XGBClassifier):
        # check whether params exist and using xgboost defaults if not using defaults
        # from https://github.com/dmlc/xgboost/blob/master/python-package/xgboost/sklearn.py
        # and here: https://xgboost.readthedocs.io/en/stable/parameter.html
        n_estimators = int(model.n_estimators) if model.n_estimators else 100
        max_depth = float(model.max_depth) if model.max_depth else 6
        min_child_weight = (
            float(model.min_child_weight) if model.min_child_weight else 1.0
        )

        if (
            (max_depth > 3.5 and 3.5 < n_estimators <= 12.5 and min_child_weight <= 1.5)
            or (max_depth > 3.5 and n_estimators > 12.5 and min_child_weight <= 3)
            or (max_depth > 3.5 and n_estimators > 62.5 and 3 < min_child_weight <= 6)
        ):
            unnecessary_risk = 1
    return unnecessary_risk


def get_tree_parameter_count(dtree: DecisionTreeClassifier) -> int:
    """Read the tree structure and return the number of learned parameters."""
    n_nodes = dtree.tree_.node_count
    left = dtree.tree_.children_left
    right = dtree.tree_.children_right
    is_leaf = np.zeros(n_nodes, dtype=int)
    for node_id in range(n_nodes):
        if left[node_id] == right[node_id]:
            is_leaf[node_id] = 1
    n_leaves = is_leaf.sum()

    # degrees of freedom
    n_internal_nodes = n_nodes - n_leaves
    n_params = 2 * n_internal_nodes  # feature id and threshold
    n_params += n_leaves * (dtree.n_classes_ - 1)  # probability distribution
    return n_params


def get_model_param_count(model: BaseEstimator) -> int:
    """Return the number of trained parameters in a model."""
    n_params = 0

    if isinstance(model, DecisionTreeClassifier):
        n_params = get_tree_parameter_count(model)

    elif isinstance(model, RandomForestClassifier):
        for member in model.estimators_:
            n_params += get_tree_parameter_count(member)

    elif isinstance(model, AdaBoostClassifier):
        try:  # sklearn v1.2+
            base = model.estimator
        except AttributeError:  # sklearn version <1.2
            base = model.base_estimator
        if isinstance(base, DecisionTreeClassifier):
            for member in model.estimators_:
                n_params += get_tree_parameter_count(member)

    # TO-DO define these for xgb, logistic regression, SVC and others
    elif isinstance(model, XGBClassifier):
        df = model.get_booster().trees_to_dataframe()
        n_trees = df["Tree"].max()
        total = len(df)
        n_leaves = len(df[df.Feature == "Leaf"])
        # 2 per internal node, one per clas in leaves, one weight per tree
        n_params = 2 * (total - n_leaves) + (model.n_classes_ - 1) * n_leaves + n_trees

    elif isinstance(model, MLPClassifier):
        weights = model.coefs_  # dtype is list of numpy.ndarrays
        biasses = model.intercepts_  # dtype is list of numpy.ndarrays
        n_params = sum(a.size for a in weights) + sum(a.size for a in biasses)

    else:
        pass

    return n_params


class StructuralAttack(Attack):
    """Structural attacks based on the static structure of a model."""

    # pylint: disable=too-many-instance-attributes

    def __init__(  # pylint: disable = too-many-arguments
        self,
        attack_config_json_file_name: str = None,
        risk_appetite_config: str = "default",
        target_path: str = None,
        output_dir: str = "outputs_structural",
        report_name: str = "report_structural",
    ) -> None:
        """Construct an object to execute a structural attack.

        Parameters
        ----------
        attack_config_json_file_name : str
            Name of a JSON file containing an attack configuration.
        risk_appetite_config : str
            Path to yaml file specifying TRE risk appetite.
        target_path : str
            Path to the saved trained target model and target data.
        output_dir : str
            Name of a directory to write outputs.
        report_name : str
            Name of the pdf and json output reports.
        """
        super().__init__()
        logger = logging.getLogger("structural_attack")
        self.target: Target = None
        self.target_path = target_path
        self.attack_config_json_file_name = attack_config_json_file_name
        # disclosure risk
        self.k_anonymity_risk = 0
        self.DoF_risk = 0
        self.unnecessary_risk = 0
        self.class_disclosure_risk = 0
        self.lowvals_cd_risk = 0
        self.metadata = {}
        # make dummy acro object and use it to extract risk appetite
        myacro = ACRO(risk_appetite_config)
        self.risk_appetite_config = risk_appetite_config
        self.THRESHOLD = myacro.config["safe_threshold"]
        self.DOF_THRESHOLD = myacro.config["safe_dof_threshold"]
        logger.info(
            "Thresholds for count %i and Dof %i", self.THRESHOLD, self.DOF_THRESHOLD
        )
        del myacro
        if self.attack_config_json_file_name is not None:
            self._update_params_from_config_file()

        # metrics
        self.attack_metrics = [
            "DoF_risk",
            "k_anonymity_risk",
            "class_disclosure_risk",
            "lowvals_cd_risk",
            "unnecessary_risk",
        ]
        self.yprobs = []

        # paths for reporting
        self.output_dir = output_dir
        self.report_name = report_name

    def __str__(self) -> str:
        """Return the name of the attack."""
        return "Structural attack"

    def attack(self, target: Target) -> None:
        """Run structural attack.

        To be used when code has access to Target class and trained target model.

        Parameters
        ----------
        target : attacks.target.Target
            target as a Target class object
        """
        self.target = target
        if target.model is None:
            errstr = (
                "cannot currently call StructuralAttack.attack() "
                "unless the target contains a trained model"
            )
            raise NotImplementedError(errstr)

        # get proba values for training data
        x = self.target.X_train
        y = self.target.y_train
        assert x.shape[0] == len(y), "length mismatch between trainx and trainy"
        self.yprobs = self.target.model.predict_proba(x)

        # only equivalence classes and membership once as potentially slow
        if isinstance(target.model, DecisionTreeClassifier):
            equiv = self.dt_get_equivalence_classes()
        else:
            equiv = self.get_equivalence_classes()
        equiv_classes = equiv[0]
        equiv_counts = equiv[1]
        equiv_members = equiv[2]
        errstr = "len mismatch between equiv classes and "
        assert len(equiv_classes) == len(equiv_counts), errstr + "counts"
        assert len(equiv_classes) == len(equiv_members), errstr + "membership"

        # now assess the risk
        # Degrees of Freedom
        n_params = get_model_param_count(target.model)
        residual_dof = self.target.X_train.shape[0] - n_params
        self.DoF_risk = 1 if residual_dof < self.DOF_THRESHOLD else 0

        # k-anonymity
        mink = np.min(np.array(equiv_counts))
        self.k_anonymity_risk = 1 if mink < self.THRESHOLD else 0

        # unnecessary risk arising from poor hyper-parameter combination.
        self.unnecessary_risk = get_unnecessary_risk(self.target.model)

        # class disclosure
        freqs = np.zeros(equiv_classes.shape)
        for group in range(freqs.shape[0]):
            freqs = equiv_classes[group] * equiv_counts[group]
        self.class_disclosure_risk = np.any(freqs < self.THRESHOLD).astype(int)
        freqs[freqs == 0] = 100
        self.lowvals_cd_risk = np.any(freqs < self.THRESHOLD).astype(int)

    def dt_get_equivalence_classes(self) -> tuple:
        """Get details of equivalence classes based on white box inspection."""
        destinations = self.target.model.apply(self.target.X_train)
        ret_tuple = np.unique(destinations, return_counts=True)
        leaves = ret_tuple[0]
        counts = ret_tuple[1]
        members = []
        for leaf in leaves:
            ingroup = np.asarray(destinations == leaf).nonzero()[0]
            members.append(ingroup)

        equiv_classes = np.zeros((len(leaves), self.target.model.n_classes_))
        for group in range(len(leaves)):
            sample_id = members[group][0]
            sample = self.target.X_train[sample_id]
            proba = self.target.model.predict_proba(sample.reshape(1, -1))
            equiv_classes[group] = proba
        return [equiv_classes, counts, members]

    def get_equivalence_classes(self) -> tuple:
        """Get details of equivalence classes based on predicted probabilities."""
        uniques = np.unique(self.yprobs, axis=0, return_counts=True)
        equiv_classes = uniques[0]
        equiv_counts = uniques[1]
        members = []
        for prob_vals in equiv_classes:
            ingroup = np.unique(np.asarray(self.yprobs == prob_vals).nonzero()[0])
            members.append(ingroup)
        return [equiv_classes, equiv_counts, members]

    def _get_global_metrics(self, attack_metrics: list) -> dict:
        """Get dictionary summarising metrics from a metric list.

        Parameters
        ----------
        attack_metrics : List
            list of attack metrics to be reported.

        Returns
        -------
        global_metrics : Dict
            Dictionary of summary metrics
        """
        global_metrics = {}
        if attack_metrics is not None and len(attack_metrics) != 0:
            global_metrics["DoF_risk"] = self.DoF_risk
            global_metrics["k_anonymity_risk"] = self.k_anonymity_risk
            global_metrics["class_disclosure_risk"] = self.class_disclosure_risk
            global_metrics["unnecessary_risk"] = self.unnecessary_risk
            global_metrics["lowvals_cd_risk"] = self.lowvals_cd_risk

        return global_metrics

    def _construct_metadata(self):
        """Construct the metadata object, after attacks."""
        self.metadata = {}
        # Store all args
        self.metadata["experiment_details"] = {}
        self.metadata["experiment_details"] = self.get_params()
        self.metadata["attack"] = str(self)
        # Global metrics
        self.metadata["global_metrics"] = self._get_global_metrics(self.attack_metrics)

    def _get_attack_metrics_instances(self) -> dict:
        """Construct the metadata object, after attacks."""
        attack_metrics_experiment = {}
        attack_metrics_instances = {}
        attack_metrics_experiment["attack_instance_logger"] = attack_metrics_instances
        attack_metrics_experiment["DoF_risk"] = self.DoF_risk
        attack_metrics_experiment["k_anonymity_risk"] = self.k_anonymity_risk
        attack_metrics_experiment["class_disclosure_risk"] = self.class_disclosure_risk
        attack_metrics_experiment["unnecessary_risk"] = self.unnecessary_risk
        attack_metrics_experiment["lowvals_cd_risk"] = self.lowvals_cd_risk
        return attack_metrics_experiment

    def make_report(self) -> dict:
        """Create output dict and generate pdf and json if filenames are given."""
        output = {}
        output["log_id"] = str(uuid.uuid4())
        output["log_time"] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        self._construct_metadata()
        output["metadata"] = self.metadata
        output["attack_experiment_logger"] = self._get_attack_metrics_instances()
        report_dest = os.path.join(self.output_dir, self.report_name)
        json_attack_formatter = GenerateJSONModule(report_dest + ".json")
        json_report = report.create_json_report(output)
        json_attack_formatter.add_attack_output(json_report, "StructuralAttack")
        return output


def _run_attack(args) -> None:
    """Initialise class and run attack."""
    attack_obj = StructuralAttack(
        risk_appetite_config=args.risk_appetite_config,
        target_path=args.target_path,
        output_dir=args.output_dir,
        report_name=args.report_name,
    )

    target = Target()
    target.load(attack_obj.target_path)
    attack_obj.attack(target)
    _ = attack_obj.make_report()


def _run_attack_from_configfile(args) -> None:
    """Initialise class and run attack using config file."""
    attack_obj = StructuralAttack(
        attack_config_json_file_name=str(args.attack_config_json_file_name),
        target_path=str(args.target_path),
    )
    target = Target()
    target.load(attack_obj.target_path)
    attack_obj.attack(target)
    _ = attack_obj.make_report()


def main() -> None:
    """Parse arguments and invoke relevant method."""
    logger = logging.getLogger("main")
    parser = argparse.ArgumentParser(description="Perform a structural  attack")

    subparsers = parser.add_subparsers()

    attack_parser = subparsers.add_parser("run-attack")

    attack_parser.add_argument(
        "--output-dir",
        type=str,
        action="store",
        dest="output_dir",
        default="output_structural",
        required=False,
        help=("Directory name where output files are stored. Default = %(default)s."),
    )

    attack_parser.add_argument(
        "--report-name",
        type=str,
        action="store",
        dest="report_name",
        default="report_structural",
        required=False,
        help=(
            """Filename for the pdf and json report outputs. Default = %(default)s.
            Code will append .pdf and .json"""
        ),
    )

    attack_parser.add_argument(
        "--risk-appetite-filename",
        action="store",
        type=str,
        default="default",
        required=False,
        dest="risk_appetite_config",
        help=(
            """provide the name of the dataset-specific risk appetite filename
            using --risk-appetite-filename Default = %(default)s"""
        ),
    )

    attack_parser.add_argument(
        "--target-path",
        action="store",
        type=str,
        default=None,
        required=False,
        dest="target_path",
        help=(
            """Provide the path to the stored target usinmg
             --target-path option. Default = %(default)f"""
        ),
    )

    attack_parser.set_defaults(func=_run_attack)

    attack_parser_config = subparsers.add_parser("run-attack-from-configfile")
    attack_parser_config.add_argument(
        "-j",
        "--attack-config-json-file-name",
        action="store",
        required=True,
        dest="attack_config_json_file_name",
        type=str,
        default="config_structural_cmd.json",
        help=(
            "Name of the .json file containing details for the run. Default = %(default)s"
        ),
    )

    attack_parser_config.add_argument(
        "-t",
        "--attack-target-folder-path",
        action="store",
        required=True,
        dest="target_path",
        type=str,
        default="structural_target",
        help=(
            """Name of the target directory to load the trained target model and the target data.
            Default = %(default)s"""
        ),
    )

    attack_parser_config.set_defaults(func=_run_attack_from_configfile)

    args = parser.parse_args()

    try:
        args.func(args)
    except AttributeError as e:  # pragma:no cover
        logger.error("Invalid command. Try --help to get more details")
        logger.error(e)


if __name__ == "__main__":  # pragma:no cover
    main()



"""Class to evaluate metric for fail fast option."""

from __future__ import annotations

from typing import Any


class FailFast:
    """Class to check attack being successful or not for a given metric.

    Note: An object of a FailFast is stateful and instance members
    (success_count and fail_count) will preserve values across repetitions for
    a test. For the new test a new object will require to be instantiated.
    """

    def __init__(self, attack_obj: Any) -> None:
        self.metric_name = attack_obj.attack_metric_success_name
        self.metric_success_thresh = attack_obj.attack_metric_success_thresh
        self.comp_type = attack_obj.attack_metric_success_comp_type
        self.success_count = 0
        self.fail_count = 0

    def check_attack_success(self, metric_dict: dict) -> bool:
        """Check if attack was successful for a given metric.

        Parameters
        ----------
        metric_dict : dict
            Dictionary with all computed metric values.

        Returns
        -------
        success_status : bool
            Boolean value based on the comparison for a given threshold.

        Notes
        -----
        If value of a given metric value has a value meeting the threshold
        based on the comparison type returns true otherwise it returns false.
        This function also counts how many times the attack was successful
        (i.e. true) and how many times it was not successful (i.e. false).
        """
        metric_value = metric_dict[self.metric_name]
        success_status = False
        if self.comp_type == "lt":
            success_status = bool(metric_value < self.metric_success_thresh)
        elif self.comp_type == "lte":
            success_status = bool(metric_value <= self.metric_success_thresh)
        elif self.comp_type == "gt":
            success_status = bool(metric_value > self.metric_success_thresh)
        elif self.comp_type == "gte":
            success_status = bool(metric_value >= self.metric_success_thresh)
        elif self.comp_type == "eq":
            success_status = bool(metric_value == self.metric_success_thresh)
        elif self.comp_type == "not_eq":
            success_status = bool(metric_value != self.metric_success_thresh)
        if success_status:
            self.success_count += 1
        else:
            self.fail_count += 1
        return success_status

    def get_success_count(self) -> int:
        """Return a count of attack being successful."""
        return self.success_count

    def get_fail_count(self) -> int:
        """Return a count of attack being not successful."""
        return self.fail_count

    def get_attack_summary(self) -> dict:
        """Return a dict of counts of attack being successful and not successful."""
        summary = {}
        summary["success_count"] = self.success_count
        summary["fail_count"] = self.fail_count
        return summary

    def check_overall_attack_success(self, attack_obj: Any) -> bool:
        """Return true if attack is successful for a given success count threshold."""
        overall_success_status = False
        if self.success_count >= attack_obj.attack_metric_success_count_thresh:
            overall_success_status = True
        return overall_success_status



"""Base class for an attack object."""

import inspect
import json

from aisdc.attacks.target import Target


class Attack:
    """Base (abstract) class to represent an attack."""

    def __init__(self) -> None:
        self.attack_config_json_file_name = None

    def attack(self, target: Target) -> None:
        """Run an attack."""
        raise NotImplementedError

    def __str__(self) -> str:
        """Return the string representation of an attack."""
        raise NotImplementedError

    def _update_params_from_config_file(self) -> None:
        """Read a configuration file and load it into a dictionary object."""
        with open(self.attack_config_json_file_name, encoding="utf-8") as f:
            config = json.loads(f.read())
        for key, value in config.items():
            setattr(self, key, value)

    @classmethod
    def _get_param_names(cls) -> list[str]:
        """Get parameter names."""
        init_signature = inspect.signature(cls.__init__)
        return [
            p.name
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]

    def get_params(self) -> dict:
        """Get parameters for this attack.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        out: dict = {}
        for key in self._get_param_names():
            out[key] = getattr(self, key)
        return out



"""Code for automatic report generation."""

import abc
import json
import os
from typing import Any

import numpy as np
import pylab as plt
from fpdf import FPDF
from pypdf import PdfWriter

# Adds a border to all pdf cells of set to 1 -- useful for debugging
BORDER = 0

DISPLAY_METRICS = [
    "AUC",
    "ACC",
    "Advantage",
    "FDIF01",
    "PDIF01",
    "TPR@0.1",
    "TPR@0.01",
    "TPR@0.001",
    "TPR@1e-05",
]

MAPPINGS = {"PDIF01": lambda x: np.exp(-x)}

INTRODUCTION = (
    "This report provides a summary of a series of simulated attack experiments performed "
    "on the model outputs provided. An attack model is trained to attempt to distinguish "
    "between outputs from training (in-sample) and testing (out-of-sample) data. The metrics "
    "below describe the success of this classifier. A successful classifier indicates that the "
    "original model is unsafe and should not be allowed to be released from the TRE.\n"
    "In particular, the simulation splits the data provided into test and train sets (each will "
    "in- and out-of-sample examples). The classifier is trained on the train set and evaluated "
    "on the test set. This is repeated with different train/test splits a user-specified number "
    "of times.\n"
    "To help place the results in context, the code may also have run a series of baseline "
    "experiments. In these, random model outputs for hypothetical in- and out-of-sample data are "
    "generated with identical statistical properties. In these baseline cases, there is no signal "
    "that an attacker could leverage and therefore these values provide a baseline against "
    "which the actual values can be compared.\n"
    "For some metrics (FDIF and AUC), we are able to compute p-values. In each case, shown below "
    "(in the Global metrics sections) is the number of repetitions that exceeded the p-value "
    "threshold both without, and with correction for multiple testing (Benjamini-Hochberg "
    "procedure).\n"
    "ROC curves for all real (red) and dummy (blue) repetitions are provided. These are shown in "
    "log space (as reommended here [ADD URL]) to emphasise the region in which risk is highest -- "
    "the bottom left (are high true positive rates possible with low false positive rates).\n"
    "A description of the metrics and how to interpret them within the context of an attack is "
    "given below."
)

LOGROC_CAPTION = (
    "This plot shows the False Positive Rate (x) versus the True Positive Rate (y). "
    "The axes are in log space enabling us to focus on areas where the False Positive Rate is low "
    "(left hand area). Curves above the y = x line (black dashes) in this region represent a "
    "disclosure risk as an attacker can obtain many more true than false positives. "
    "The solid coloured lines show the curves for the attack simulations with the true model "
    "outputs. The lighter grey lines show the curves for randomly generated outputs with "
    "no structure (i.e. in- and out-of- sample predictions are generated from the same "
    "distributions. Solid curves consistently higher than the grey curves in the left hand "
    "part of the plot are a sign of concern."
)

GLOSSARY = {
    "AUC": "Area Under the ROC curve",
    "True Positive Rate (TPR)": (
        "The true positive rate is the number of True Positives that are predicted as positive as "
        "a proportion of the total number of positives. If an attacker has N examples that were "
        "actually in the training set, the TPR is the proportion of these that they predict as "
        "being in the training set."
    ),
    "ACC": "The proportion of predictions that the attacker makes that are correct.",
}


class NumpyArrayEncoder(json.JSONEncoder):
    """Json encoder that can cope with numpy arrays."""

    def default(self, o: Any):
        """If an object is an np.ndarray, convert to list."""
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, np.int64):
            return int(o)
        if isinstance(o, np.int32):
            return int(o)
        if isinstance(o, abc.ABCMeta):
            return str(o)
        return json.JSONEncoder.default(self, o)


def _write_dict(
    pdf: FPDF, input_dict: dict, indent: int = 0, border: int = BORDER
) -> None:
    """Write a dictionary to the pdf."""
    for key, value in input_dict.items():
        pdf.set_font("arial", "B", 14)
        pdf.cell(75, 5, key, border, 1, "L")
        pdf.cell(indent, 0)
        pdf.set_font("arial", "", 12)
        pdf.multi_cell(150, 5, value, border, "L")
        pdf.ln(h=5)


def title(
    pdf: FPDF,
    text: str,
    border: int = BORDER,
    font_size: int = 24,
    font_style: str = "B",
) -> None:
    """Write a title block."""
    pdf.set_font("arial", font_style, font_size)
    pdf.ln(h=5)
    pdf.cell(0, 0, text, border, 1, "C")
    pdf.ln(h=5)


def subtitle(  # pylint: disable = too-many-arguments
    pdf: FPDF,
    text: str,
    indent: int = 10,
    border: int = BORDER,
    font_size: int = 12,
    font_style: str = "B",
) -> None:
    """Write a subtitle block."""
    pdf.cell(indent, border=border)
    pdf.set_font("arial", font_style, font_size)
    pdf.cell(75, 10, text, border, 1)


def line(  # pylint: disable = too-many-arguments
    pdf: FPDF,
    text: str,
    indent: int = 0,
    border: int = BORDER,
    font_size: int = 11,
    font_style: str = "",
    font: str = "arial",
) -> None:
    """Write a standard block."""
    if indent > 0:
        pdf.cell(indent, border=border)
    pdf.set_font(font, font_style, font_size)
    pdf.multi_cell(0, 5, text, border, 1)


def _roc_plot_single(metrics: dict, save_name: str) -> None:
    """Create a roc_plot for a single experiment."""
    plt.figure()
    plt.plot([0, 1], [0, 1], "k--")
    plt.plot(metrics["fpr"], metrics["tpr"], "r", linewidth=2)
    plt.xscale("log")
    plt.yscale("log")
    plt.grid()
    plt.ylabel("True Positive Rate")
    plt.xlabel("False Positive Rate")
    plt.tight_layout()
    plt.savefig(save_name)


def _roc_plot(metrics: dict, dummy_metrics: list, save_name: str) -> None:
    """Create a roc plot for multiple repetitions."""
    plt.figure()
    plt.plot([0, 1], [0, 1], "k--")
    do_dummy = bool(dummy_metrics)

    # Compute average ROC
    base_fpr = np.linspace(0, 1, 1000)
    all_tpr = np.zeros((len(metrics), len(base_fpr)), float)
    for i, metric_set in enumerate(metrics):
        all_tpr[i, :] = np.interp(base_fpr, metric_set["fpr"], metric_set["tpr"])

    if do_dummy:
        all_tpr_dummy = np.zeros((len(dummy_metrics), len(base_fpr)), float)
        for i, metric_set in enumerate(dummy_metrics):
            all_tpr_dummy[i, :] = np.interp(
                base_fpr, metric_set["fpr"], metric_set["tpr"]
            )

        for _, metric_set in enumerate(dummy_metrics):
            plt.plot(
                metric_set["fpr"],
                metric_set["tpr"],
                color="lightsteelblue",
                linewidth=0.5,
                alpha=0.5,
            )

    for _, metric_set in enumerate(metrics):
        plt.plot(
            metric_set["fpr"], metric_set["tpr"], color="lightsalmon", linewidth=0.5
        )

    tpr_mu = all_tpr.mean(axis=0)
    plt.plot(base_fpr, tpr_mu, "r")

    if do_dummy:
        dummy_mu = all_tpr_dummy.mean(axis=0)
        plt.plot(base_fpr, dummy_mu, "b")

    plt.xscale("log")
    plt.yscale("log")
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.tight_layout()
    plt.grid()
    plt.savefig(save_name)


def create_mia_report(attack_output: dict) -> FPDF:
    """Make a worst case membership inference report.

    Parameters
    ----------
    attack_output : dict
        Dictionary with the following items:

        metadata : dict
            Dictionary of metadata.
        attack_experiment_logger : dict
            List of metrics as dictionary items for an experiment.
        dummy_attack_experiment_logger : dict
            List of metrics as dictionary items across dummy experiments.

    Returns
    -------
    pdf : fpdf.FPDF
        fpdf document object
    """
    do_dummy = False
    dummy_metrics = []
    mia_metrics = [
        v
        for _, v in attack_output["attack_experiment_logger"][
            "attack_instance_logger"
        ].items()
    ]
    metadata = attack_output["metadata"]

    dest_log_roc = (
        os.path.join(
            metadata["experiment_details"]["output_dir"],
            metadata["experiment_details"]["report_name"],
        )
        + "_log_roc.png"
    )
    _roc_plot(mia_metrics, dummy_metrics, dest_log_roc)

    pdf = FPDF()
    pdf.add_page()
    pdf.set_xy(0, 0)
    title(pdf, "WorstCase MIA attack result report")
    subtitle(pdf, "Introduction")
    line(pdf, INTRODUCTION)
    subtitle(pdf, "Experiment summary")
    for key, value in metadata["experiment_details"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    subtitle(pdf, "Global metrics")
    for key, value in metadata["global_metrics"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    if do_dummy:
        subtitle(pdf, "Baseline global metrics")
        for key, value in metadata["baseline_global_metrics"].items():
            line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")

    subtitle(pdf, "Metrics")
    line(
        pdf,
        "The following show summaries of the attack metrics over the repetitions",
        font="arial",
    )
    for metric in DISPLAY_METRICS:
        vals = np.array([m[metric] for m in mia_metrics])
        if metric in MAPPINGS:
            vals = np.array([MAPPINGS[metric](v) for v in vals])
        text = (
            f"{metric:>12} mean = {vals.mean():.2f}, var = {vals.var():.4f}, "
            f"min = {vals.min():.2f}, max = {vals.max():.2f}"
        )
        line(pdf, text, font="courier")

    if do_dummy:
        subtitle(pdf, "Baseline metrics")
        line(
            pdf,
            (
                "The following show summaries of the attack metrics over the "
                "repetitions where there is no statistical difference between "
                "predictions in the training and test sets. Simulation was done "
                "with training and test set sizes equal to the real ones"
            ),
            font="arial",
        )
        for metric in DISPLAY_METRICS:
            vals = np.array([m[metric] for m in dummy_metrics])
            if metric in MAPPINGS:
                vals = np.array([MAPPINGS[metric](v) for v in vals])
            text = (
                f"{metric:>12} mean = {vals.mean():.2f}, var = {vals.var():.4f}, "
                f"min = {vals.min():.2f}, max = {vals.max():.2f}"
            )
            line(pdf, text, font="courier")

    _add_log_roc_to_page(dest_log_roc, pdf)
    line(pdf, LOGROC_CAPTION)

    pdf.add_page()
    title(pdf, "Glossary")
    _write_dict(pdf, GLOSSARY)

    return pdf


def add_output_to_pdf(report_dest: str, pdf_report: FPDF, attack_type: str) -> None:
    """Create pdf and append contents if it already exists."""
    if os.path.exists(report_dest + ".pdf"):
        old_pdf = report_dest + ".pdf"
        new_pdf = report_dest + "_new.pdf"
        pdf_report.output(new_pdf)
        merger = PdfWriter()
        for pdf in [old_pdf, new_pdf]:
            merger.append(pdf)
        merger.write(old_pdf)
        merger.close()
        os.remove(new_pdf)
    else:
        pdf_report.output(report_dest + ".pdf")
    if attack_type in ("WorstCaseAttack", "LikelihoodAttack"):
        path = report_dest + "_log_roc.png"
        if os.path.exists(path):
            os.remove(path)
    elif attack_type == "AttributeAttack":
        path = report_dest + "_cat_frac.png"
        if os.path.exists(path):
            os.remove(path)

        path = report_dest + "_cat_risk.png"
        if os.path.exists(path):
            os.remove(path)


def _add_log_roc_to_page(log_roc: str = None, pdf_obj: FPDF = None) -> None:
    if log_roc is not None:
        pdf_obj.add_page()
        subtitle(pdf_obj, "Log ROC")
        pdf_obj.image(log_roc, x=None, y=None, w=0, h=140, type="", link="")
        pdf_obj.set_font("arial", "", 12)


def create_json_report(output: dict) -> None:
    """Create a report in json format for injestion by other tools."""
    # Initial work, just dump mia_metrics and dummy_metrics into a json structure
    return json.dumps(output, cls=NumpyArrayEncoder)


def create_lr_report(output: dict) -> FPDF:
    """Make a lira membership inference report.

    Parameters
    ----------
    output : dict
        Dictionary with the following items:

        metadata : dict
            Dictionary of metadata.

        attack_experiment_logger : dict
            List of metrics as dictionary items for an experiments.
            In case of LiRA attack scenario, this will have dictionary items of
            `attack_instance_logger` that will have a single metrics dictionary.

    Returns
    -------
    pdf : fpdf.FPDF
        fpdf document object
    """
    mia_metrics = [
        v
        for _, v in output["attack_experiment_logger"]["attack_instance_logger"].items()
    ][0]
    metadata = output["metadata"]
    dest_log_roc = (
        os.path.join(
            metadata["experiment_details"]["output_dir"],
            metadata["experiment_details"]["report_name"],
        )
        + "_log_roc.png"
    )
    _roc_plot_single(mia_metrics, dest_log_roc)
    pdf = FPDF()
    pdf.add_page()
    pdf.set_xy(0, 0)
    title(pdf, "Likelihood Ratio Attack Report")
    subtitle(pdf, "Introduction")
    subtitle(pdf, "Metadata")
    for key, value in metadata["experiment_details"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    for key, value in metadata["global_metrics"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    subtitle(pdf, "Metrics")
    sub_metrics_dict = {
        key: val for key, val in mia_metrics.items() if isinstance(val, float)
    }
    for key, value in sub_metrics_dict.items():
        val = MAPPINGS[key](value) if key in MAPPINGS else value
        line(pdf, f"{key:>30s}: {val:.4f}", font="courier")

    pdf.add_page()
    subtitle(pdf, "ROC Curve")
    pdf.image(dest_log_roc, x=None, y=None, w=0, h=140, type="", link="")
    return pdf
